/*
 * Created on Mar 30, 2005
 *
 * $Id: CFG.java,v 1.3 2005/04/27 13:56:08 dragos Exp $
 */
package einsc.backend.cfg;

import java.util.*;
import einsc.backend.*;

/**
 * Control-flow graph (CFG). A CFG is composed of nodes, which are
 * basic blocks containing non-jumping instructions, and edges, which
 * are labelled with jump instructions.
 */
public class CFG {
    public final String name;

	private LinkedList<Node> nodes = new LinkedList<Node>();
	private ArrayList<Node> labels = new ArrayList<Node>();
	private HashMap<Node, List<Edge>> edges = new HashMap<Node, List<Edge>>();
	private HashMap<Node, List<Edge>> revEdges = new HashMap<Node, List<Edge>>();

	public CFG(String name) {
		this.name = name;
	}

	public class Node {
		/** Instructions composing this node (basic block) */
		private Code code = new Code();

		/** Label number associated with this node */
		public final int label;

		public Node() {
			label = labels.size();
			labels.add(this);
		}

		public void emit(Instruction i) {
			code.emit(i);
		}

		public void emit(int opcode) {
			code.emit(opcode);
		}

		public void emit(int opcode, int c) {
			code.emit(opcode, c);
		}

		public void emit(int opcode, String c) {
			code.emit(opcode, c);
		}

		public void emit(int opcode, int a, int c) {
			code.emit(opcode, a, c);
		}

		public void emit(int opcode, int a, String c) {
			code.emit(opcode, a, c);
		}

		public void emit(int opcode, int a, int b, int c) {
			code.emit(opcode, a, b, c);
		}

		public int size() {
			return code.size();
		}
		
        /** Return an iterator over the instructions of this node.<p>
         *  It contains all the instructions associated to outer
         *  edges of this node, if any. It is a bidirectional iterator.
         */
        public ListIterator<Instruction> iterator() {
        	return new InstructionIterator();
        }
        
        /** Iterator over all instructions in this node. <p>
         * It implmements only bidirectional navigation, no changes to
         * the underlying data is allowed.
         */
        public class InstructionIterator implements ListIterator<Instruction> {
        	ListIterator<Instruction> blockIterator;
            ListIterator<Edge> edgeIterator;
            boolean    onEdges;
            
            public InstructionIterator() {
            	blockIterator = code.instructions().listIterator();
                edgeIterator  = getEdgesFrom(Node.this).listIterator();
                onEdges       = false;
            }
            
			public boolean hasNext() {
                if (!onEdges) {
                	if (blockIterator.hasNext())
                        return true;
                    else {
                    	onEdges = true;
                        return edgeIterator.hasNext();
                    }
                } else
                    return edgeIterator.hasNext();
			}

			public Instruction next() {
                if (!onEdges) {
                    if (blockIterator.hasNext())
                        return blockIterator.next();
                    else {
                        onEdges = true;
                        return edgeIterator.next().branch;
                    }
                } else
                    return edgeIterator.next().branch;
			}

			public void remove() {
                throw new UnsupportedOperationException();
			}

			public boolean hasPrevious() {
                if (onEdges) {
                    if (edgeIterator.hasPrevious())
                        return true;
                    else {
                        onEdges = false;
                        return blockIterator.hasPrevious();
                    }
                } else
                    return blockIterator.hasPrevious();
			}

			public Instruction previous() {
                if (onEdges) {
                	if (edgeIterator.hasPrevious())
                        return edgeIterator.previous().branch;
                    else {
                    	onEdges = false;
                        return blockIterator.previous();
                    }
                } else
                    return blockIterator.previous();
			}

			public int nextIndex() {
                throw new UnsupportedOperationException();
			}

			public int previousIndex() {
                throw new UnsupportedOperationException();
			}

			public void set(Instruction arg0) {
                throw new UnsupportedOperationException();
			}

			public void add(Instruction arg0) {
                throw new UnsupportedOperationException();
			}
        	
        }
	
        /** Return an iterator for the instructions inside the node. */
        public ListIterator<Instruction> blockIterator() {
        	return Collections.unmodifiableList(code.instructions()).listIterator();
        }

        /** Return an iterator for the instructions inside the node. */
        public ListIterator<Instruction> blockIterator(boolean startAtEnd) {
            if (startAtEnd)
            	return Collections.unmodifiableList(code.instructions()).listIterator(code.instructions().size());
            else
                return Collections.unmodifiableList(code.instructions()).listIterator();
        }

        /** Replace the code of this node with the new code. */
        public void replaceCode(Code c) {
        	this.code = c;
        }
    }

	public class Edge {
		public final Node from, to;

		public final Instruction branch;

		public Edge(Node from, Node to, int opcode, int reg) {
			this.from = from;
			this.to = to;
			this.branch = Instruction.make(opcode, reg, name + "_" + to.label);
		}
	}

	public Node addNewNode() {
		Node newNode = new Node();
		nodes.add(newNode);
		return newNode;
	}

	public Node getNode(int label) {
		return (Node) labels.get(label);
	}

	public List<Node> getNodes() {
		return Collections.unmodifiableList(nodes);
	}

	public Edge addNewEdge(Node from, Node to, int opcode, int reg) {
		Edge newEdge = new Edge(from, to, opcode, reg);

		List<Edge> fromL = edges.get(from);
		if (fromL == null) {
			fromL = new LinkedList<Edge>();
			edges.put(from, fromL);
		}
		fromL.add(newEdge);
        
        assert fromL.size() < 3 : "Too many outgoing edges.";

		List<Edge> toL = revEdges.get(to);
		if (toL == null) {
			toL = new LinkedList<Edge>();
			revEdges.put(to, toL);
		}
		toL.add(newEdge);

		return newEdge;
	}

    /** Return the last instruction of this node. 
     *  <p>If this node has no successors, it returns the last one in the
     *  code list. If it has successors, it returns the instruction
     *  attached to the first edge in the list. 
     */
	public Instruction getLastInstruction(Node node) {
		List<Edge> edgesFrom = getEdgesFrom(node);
		if (edgesFrom.isEmpty()) {
			List<Instruction> instrs = node.code.instructions();
			if (instrs.isEmpty())
				return null;
			else
				return /*(Instruction)*/ instrs.get(instrs.size() - 1);
		} else
			return ((Edge) edgesFrom.get(0)).branch;
	}

    /** Return the list of edges that start at this node.
     *  It never returns null.  
     */
	public List<Edge> getEdgesFrom(Node node) {
		List<Edge> edgesFrom = edges.get(node);
		if (edgesFrom != null)
			return Collections.unmodifiableList(edgesFrom);
		else
			return Collections.emptyList();
	}

    /**
     * Returns the list of successor nodes.
     */
	public List<Node> successors(Node node) {
		List<Node> succs = new LinkedList<Node>();
		Iterator<Edge> edgesIt = getEdgesFrom(node).iterator();
		while (edgesIt.hasNext())
			succs.add(edgesIt.next().to);
		return succs;
	}

    /**
     * Return the list of edges that enter this node.<p>
     * It never returns null.
     */
	public List<Edge> getEdgesTo(Node node) {
		List<Edge> edgesTo = revEdges.get(node);
		if (edgesTo != null)
			return Collections.unmodifiableList(edgesTo);
		else
			return Collections.emptyList();
	}

    /**
     * Return the list of predecessor nodes.
     */
	public List<Node> predecessors(Node node) {
		List<Node> preds = new LinkedList<Node>();
		Iterator<Edge> edgesIt = getEdgesTo(node).iterator();
		while (edgesIt.hasNext())
			preds.add(edgesIt.next().from);
		return preds;
	}

	/** Visit the CFG in depth-first search. */
	public void visitDFS(Visitor visitor) {
		if (!nodes.isEmpty())
			visitDFS(visitor, (Node) nodes.getFirst(), new HashSet<Node>());
	}

	private void visitDFS(Visitor visitor, Node root, HashSet<Node> visited) {
		if (!visited.contains(root)) {
			visited.add(root);
			visitor.pre(root);

			Iterator<Edge> outIt = getEdgesFrom(root).iterator();
			while (outIt.hasNext()) {
				Edge outEdge = /*(Edge)*/ outIt.next();
				visitDFS(visitor, outEdge.to, visited);
			}
			visitor.post(root);
		}
	}

	/** Visit the CFG in reverse pre-order. */
	public void visitRPO(Visitor visitor) {
		final LinkedList<Node>nodes = new LinkedList<Node>();
		visitDFS(new CFG.EmptyVisitor() {
			public void post(CFG.Node node) {
				nodes.addFirst(node);
			}
		});
		Iterator<Node> nodesIt = nodes.iterator();
		while (nodesIt.hasNext()) {
			Node n = /*(Node)*/ nodesIt.next();
			visitor.pre(n);
			visitor.post(n);
		}
	}

	/** Visitor for DFS traversal. */
	public static interface Visitor {
		/** Method called before visiting the children of the node */
		public void pre(Node node);

		/** Method called after visiting the children of the node */
		public void post(Node node);
	}

	/** Default visitor, with empty pre and post methods. */
	public static class EmptyVisitor implements Visitor {
		public void pre(Node node) {
		}

		public void post(Node node) {
		}
	}
}
